From Coq Require Import ssreflect ssrfun ssrbool.
Require Import Common.
Require Import BinNums.
Require Import stdpp.list.
Require Import stdpp.gmultiset.
Require Import stdpp.relations.
Require Import PE2.

Definition pexp_syms : relation pexp := fun p q => syms (coeqs p) (coeqs q).
Infix "≡ₓ" := pexp_syms (at level 70, no associativity).
Notation "(≡ₓ)" := pexp_syms (only parsing).
Instance pexp_syms_equivalence : Equivalence (≡ₓ).
Proof.
  rewrite /pexp_syms.
  constructor.
  - move=> *; reflexivity.
  - move=> *; by symmetry.
  - move=> *; by etransitivity.
Qed.

Inductive value_reaction : el_relation coeq :=
| unit_l : value_reaction (Coeq star costar) []
| unit_r : value_reaction (Coeq costar star) []
| pair_l : forall t u t' u',
    value_reaction (Coeq (ttens t u) (tpar t' u')) [Coeq t t'; Coeq u u']
| pair_r : forall t u t' u',
    value_reaction (Coeq (tpar t' u') (ttens t u)) [Coeq t t'; Coeq u u'].
Definition cvalue_reaction : ll_relation coeq := el_contextual value_reaction.
Inductive value_reaction_step : relation pexp :=
| ValueReaction : forall Θ Θ' ts,
    cvalue_reaction Θ Θ' -> value_reaction_step (Pexp Θ ts) (Pexp Θ' ts).
Hint Constructors value_reaction value_reaction_step.

Instance: Proper ((≡) ==> (=) ==> iff) value_reaction.
Proof.
  apply proper_sym_impl_iff_2; only 1,2: typeclasses eauto.
  move=> [t u] _ [] //= ? ? <-; by inversion 1.
Qed.
Lemma equiv_cvalue_reaction_subcommute : subcommute (≡) cvalue_reaction.
Proof. apply: soln_equiv_subcommute_el. Qed.
Lemma equiv_value_reaction_step_subcommute : subcommute (≡) value_reaction_step.
Proof.
  move=> [Θ ts] r [p_ E Value].
  case: p_ r / Value E => Ξ Ξ' _ CVR [/= E <-].
  move: (E >>> CVR) => /equiv_cvalue_reaction_subcommute [Ξ2 CVR' E'].
    by eexists.
Qed.

Definition var_is_side (x : name) : coeq -> Prop := fun '(Coeq t u) => t = x \/ u = x.
Definition reactive (x : name) (Θ : soln) (ts : list term) : Prop :=
  Exists (var_is_side x) Θ \/ x ∈ vars ts.
Instance decide_reactive : Decision (reactive x Θ ts).
Proof.
  move=> x Θ ts.
  apply or_dec; last by typeclasses eauto.
  apply Exists_dec => -[t u]; apply: or_dec.
Defined.
Instance: Proper ((=) ==> (≡) ==> iff) var_is_side.
Proof. move=> x _ <- [t u] C' [] /=; intuition. Qed.
Instance: Proper ((=) ==> (≡) ==> (≡ₚ) ==> iff) reactive.
Proof.
  move=> x _ <-.
  apply proper_sym_impl_iff_2; only 1,2: typeclasses eauto.
  move=> Θ Θ'' /soln_equiv_alt_alt [Θ' Perm All] ts ts' E_ts [InSoln | InBody];
          last by right; rewrite -E_ts //.
  move: InSoln All.
  rewrite {}Perm => /Exists_exists [C [/(elem_of_list_split _ _) [Θ'1] [Θ'2] -> Side]]
    /Forall2_app_inv_l [Θ''1] [?] [All1]
    [/(Forall2_cons_inv_l _ _) [C'] [Θ''2] [E_C] [All2] ->] ->.
  left; apply/Exists_app; right; apply/Exists_cons; left.
  rewrite -E_C //.
Qed.
Lemma var_is_side_alt : forall {x C}, var_is_side x C <-> exists t, C ≡ Coeq x t.
Proof. move=> x [t u]; split=> [[-> | ->] | [t' ->]] /=; by repeat econstructor. Qed.
Definition var_isn't_side (x : name) : coeq -> Prop := fun '(Coeq t u) => t <> x /\ u <> x.
Instance: Proper ((=) ==> (≡) ==> iff) var_isn't_side.
Proof. move=> x _ <- [t u] C' [] /=; intuition. Qed.
Lemma var_isn't_side_alt : forall {x C},
    ~var_is_side x C <-> var_isn't_side x C.
Proof. move=> x [t u] /=. tauto. Qed.
Lemma unreactive_alt : forall {x Θ ts},
    ~reactive x Θ ts <-> Forall (var_isn't_side x) Θ /\ x ∉ vars ts.
Proof.
  move=> x Θ ts.
  rewrite /reactive /not !Decidable.not_or_iff -![_ -> False]/(~_) -!Forall_Exists_neg
          !(Forall_iff _ _ _ (@var_isn't_side_alt _)) //.
Qed.
Lemma var_is_side_mult : forall {x C}, var_is_side x C -> count x C > 0.
Proof. move=> x [t u] [-> | ->]; rewrite !countE; lia. Qed.

Inductive del1 (C : coeq) : el_relation coeq :=
| Del1 : forall C', C' ≡ C -> del1 C C' [].
Definition del (C : coeq) : ll_relation coeq := el_contextual (del1 C).
Inductive reactive_transmission (x : name) : relation pexp :=
| ReactiveTransmission : forall Θ t Θ' ts,
    del (Coeq t (var x)) Θ Θ' -> reactive x Θ' ts ->
    reactive_transmission x (Pexp Θ ts) (subst t x (Pexp Θ' ts)).
Inductive unreactive_transmission (x : name) : relation pexp :=
| UnreactiveTransmission : forall Θ t Θ' ts,
    del (Coeq t (var x)) Θ Θ' -> ~reactive x Θ' ts ->
    unreactive_transmission x (Pexp Θ ts) (subst t x (Pexp Θ' ts)).
Hint Constructors reactive_transmission unreactive_transmission.
Definition transmission (x : name) : relation pexp :=
  fun p q => reactive_transmission x p q \/ unreactive_transmission x p q.
Definition reactive_transmission_step : relation pexp :=
  fun p q => exists x, reactive_transmission x p q.
Definition unreactive_transmission_step : relation pexp :=
  fun p q => exists x, unreactive_transmission x p q.
Definition transmission_step : relation pexp :=
  fun p q => exists x, transmission x p q.
Hint Unfold transmission
     reactive_transmission_step unreactive_transmission_step transmission_step.

Definition tstep : relation pexp :=
  fun p q => transmission_step p q \/ value_reaction_step p q.

Lemma transmission_step_alt :
    transmission_step ≡ᵣ reactive_transmission_step ∪ᵣ unreactive_transmission_step.
Proof. firstorder. Qed.

Lemma do_rtransmission : forall {C x Θ1 t Θ2 ts},
    C ≡ Coeq t (var x) ->
    reactive x (Θ1 ++ Θ2) ts ->
    reactive_transmission x (Pexp (Θ1 ++ [C] ++ Θ2) ts)
                          (Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts)).
Proof.
  move=> C x Θ1 t Θ2 ts E R.
  rewrite -sim_substl_app.
    by move: (el_Contextual Θ1 _ _ Θ2 (Del1 _ _ E)).
Qed.
Lemma do_rtransmission' : forall {C x Θ1 t Θ2 ts q},
    C ≡ Coeq t (var x) ->
    reactive x (Θ1 ++ Θ2) ts ->
    q = Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts) ->
    reactive_transmission x (Pexp (Θ1 ++ [C] ++ Θ2) ts) q.
Proof. move=> C x Θ1 t Θ2 ts _ E R ->; by apply: do_rtransmission. Qed.

Lemma rtransmission_l : forall {x Θ1 t Θ2 ts q},
    reactive x (Θ1 ++ Θ2) ts ->
    q = Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts) ->
    reactive_transmission x (Pexp (Θ1 ++ [Coeq (var x) t] ++ Θ2) ts) q.
Proof. move=> ? ? ? ? ? ? ? ->; apply: do_rtransmission; [constructor | done]. Qed.
Lemma rtransmission_r : forall {x Θ1 t Θ2 ts q},
    reactive x (Θ1 ++ Θ2) ts ->
    q = Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts) ->
    reactive_transmission x (Pexp (Θ1 ++ [Coeq t (var x)] ++ Θ2) ts) q.
Proof. move=> ? ? ? ? ? ? ? ->; apply: do_rtransmission; [constructor | done]. Qed.

Lemma do_transmission : forall {C x Θ1 t Θ2 ts},
    C ≡ Coeq t (var x) ->
    transmission x (Pexp (Θ1 ++ [C] ++ Θ2) ts)
                 (Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts)).
Proof.
  move=> C x Θ1 t Θ2 ts E.
  rewrite -sim_substl_app -sim_substpe_def.
  move: (el_Contextual Θ1 _ _ Θ2 (Del1 _ _ E)).
  case: (decide (reactive x (Θ1 ++ Θ2) ts)); eauto.
Qed.
Lemma do_transmission' : forall {C x Θ1 t Θ2 ts q},
    C ≡ Coeq t (var x) ->
    q = (Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts)) ->
    transmission x (Pexp (Θ1 ++ [C] ++ Θ2) ts) q.
Proof. move=> C x Θ1 t Θ2 ts _ E ->; by apply: do_transmission. Qed.

Lemma transmission_l : forall {x Θ1 t Θ2 ts q},
    q = Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts) ->
    transmission x (Pexp (Θ1 ++ [Coeq (var x) t] ++ Θ2) ts) q.
Proof. move=> ? ? ? ? ? ? ->; apply: do_transmission; constructor. Qed.
Lemma transmission_r : forall {x Θ1 t Θ2 ts q},
    q = Pexp (subst t x Θ1 ++ subst t x Θ2) (subst t x ts) ->
    transmission x (Pexp (Θ1 ++ [Coeq t (var x)] ++ Θ2) ts) q.
Proof. move=> ? ? ? ? ? ? ->; apply: do_transmission; constructor. Qed.

Lemma equiv_del_subcommute : forall {C}, subcommute (≡) (del C).
Proof.
  move=> C; unshelve eapply soln_equiv_subcommute_el.
  apply proper_sym_impl_iff_2; only 1,2: typeclasses eauto.
  move=> C' C'' E ε _ <- Del1.
  case: C' ε / Del1 E; constructor; by etransitivity.
Qed.
Lemma equiv_unreactive_transmission_subcommute :
  forall {x}, subcommute (≡) (unreactive_transmission x).
Proof.
  move=> x [Θ _ts] q [p_ E_p U].
  case: p_ q / U E_p => Ξ t Ξ' ts Del Unreactive [/= E ->].
  move: (E >>> Del) => /equiv_del_subcommute [Ξ2 Del' E'].
  eexists.
  - constructor; [done | by rewrite E'].
  - rewrite E' //.
Qed.
Corollary equiv_unreactive_transmission_step_subcommute :
  subcommute (≡) unreactive_transmission_step.
Proof.
  move=> p r [q E [x U]].
  move: (E >>> U) => /equiv_unreactive_transmission_subcommute [q' U' E'].
  exact (ex_intro _ _ U' >>> E').
Qed.
Lemma equiv_reactive_transmission_subcommute :
  forall {x}, subcommute (≡) (reactive_transmission x).
Proof.
  move=> x [Θ _ts] q [p_ E_p U].
  case: p_ q / U E_p => Ξ t Ξ' ts Del Reactive [/= E ->].
  move: (E >>> Del) => /equiv_del_subcommute [Ξ2 Del' E'].
  eexists.
  - constructor; [done | by rewrite E'].
  - rewrite E' //.
Qed.
Corollary equiv_reactive_transmission_step_subcommute :
  subcommute (≡) reactive_transmission_step.
Proof.
  move=> p r [q E [x R]].
  move: (E >>> R) => /equiv_reactive_transmission_subcommute [q' R' E'].
  exact (ex_intro _ _ R' >>> E').
Qed.
Corollary equiv_transmission_subcommute :
  forall {x}, subcommute (≡) (transmission x).
Proof.
  move=> x p r [q E [R | U]].
  - move: (E >>> R) => /equiv_reactive_transmission_subcommute [q' R' E'].
    exact (rel_inl _ R' >>> E').
  - move: (E >>> U) => /equiv_unreactive_transmission_subcommute [q' U' E'].
    exact (rel_inr _ U' >>> E').
Qed.
Corollary equiv_transmission_step_subcommute :
  subcommute (≡) transmission_step.
Proof.
  move=> p r [q E [x T]].
  move: (E >>> T) => /equiv_transmission_subcommute [q' T' E'].
  eexists; eauto.
Qed.


Inductive action : Type := silent | transmit (t : term) (x : name).
Definition apply (α : action) : pexp -> pexp :=
  match α with
  | silent => id
  | transmit t x => subst t x
  end.

Inductive coeq_tstep : action -> el_relation coeq :=
| v_coeq_tstep : forall {C Ξ}, value_reaction C Ξ -> coeq_tstep silent C Ξ
| t_coeq_tstep : forall {C t x}, C ≡ Coeq t (var x) -> coeq_tstep (transmit t x) C [].
Definition ccoeq_tstep (α : action) : ll_relation coeq := el_contextual (coeq_tstep α).
Inductive tstep' : relation pexp :=
| Tstep' : forall α {Θ Θ' ts}, ccoeq_tstep α Θ Θ' -> tstep' (Pexp Θ ts) (apply α (Pexp Θ' ts)).
Hint Constructors coeq_tstep tstep'.
Instance: Proper ((=) ==> (≡) ==> (=) ==> iff) coeq_tstep.
Proof.
  move=> α _ <-.
  apply proper_sym_impl_iff_2; only 1,2: typeclasses eauto.
  move=> [t u] _ [] //= ? ? <-; inversion 1; constructor; rewrite [Coeq u t]coeq_swap //.
Qed.

Lemma tstep_alt : forall {p q}, tstep p q <-> tstep' p q.
Proof.
  move=> p q; split.
  - move=> [[x [] [_ t Θ' ts [Θ1 _ _ Θ2 [C E]] _]] | [_ Θ' ts [Θ1 C Ξ Θ2 VR]]].
    + apply: (Tstep' (transmit t x)); by repeat constructor.
    + apply: (Tstep' (transmit t x)); by repeat constructor.
    + apply: (Tstep' silent); by repeat constructor.
  - move=> [_ _ Θ' ts [Θ1 _ _ Θ2 [C Ξ VR | C t x E]]].
    + by repeat constructor.
    + rewrite /= !substE.
      left; exists x; by apply: do_transmission'.
Qed.

Theorem strong_coeq_progress : forall {C}, weak_matched C -> exists α Ξ, coeq_tstep α C Ξ.
Proof.
  move=> [t u] [A /= [/wtht_expand [[x ->] | M_t] /wtht_expand [[y ->] | M_u]]].
  - eauto.
  - setoid_rewrite -> coeq_swap; rewrite /=; eauto.
  - eauto.
  - case: A M_t M_u => //=.
    + move=> -> ->; eauto.
    + move=> -> ->; eauto.
    + move=> A B [t_l] [t_r] [_] [_] -> [u_l] [u_r] [_] [_] ->.
      eauto.
    + move=> A B [t_l] [t_r] [_] [_] -> [u_l] [u_r] [_] [_] ->.
      eauto.
Qed.
Theorem strong_progress : forall {p Seq}, weak_has_ty p Seq -> cut_free p \/ exists q, tstep p q.
Proof.
  move=> [[| [t u] Θ] ts] /= Seq [Len [Matched _]]; [by left | right].
  move: Matched => /Forall_cons [/strong_coeq_progress [α [Ξ CTS]] _].
  eexists; apply/tstep_alt; constructor; by apply: el_beginning.
Qed.
Corollary progress : forall {p Seq}, has_ty p Seq -> cut_free p \/ exists q, tstep p q.
Proof. move=> p Seq /weak_typing_inversion'; apply: strong_progress. Qed.


Instance value_reaction_ok : subrelation value_reaction_step step.
Proof.
  move=> _ _ [_ _ ts [Θ1 C Θ Θ2 [| | t u t' u' | t u t' u']]].
  - eexists; only 1,3: reflexivity.
    eexists; last by done.
    repeat constructor.
  - eexists; first by rewrite [Coeq _ _]coeq_swap.
    eexists; last by done.
    repeat constructor.
  - eexists; only 1,3: reflexivity.
    eexists; last by done.
    repeat constructor.
  - eexists; first by rewrite [Coeq _ _]coeq_swap.
    eexists; last by done.
    repeat constructor.
Qed.

Theorem reactive_transmission_ok : forall {p q},
    linear p -> reactive_transmission_step p q -> step p q.
Proof.
  move=> p q Linear Step.
  move: Step Linear => [x [_ t _ ts [Θ1 _ _ Θ2 [C ->]] /= [/Exists_app [] | InBody]]].
  - move=> /Exists_exists [C' [/(elem_of_list_split _ _) [Θ11 [Θ12 ->]]
                               /var_is_side_alt [u ->]]] Linear.
    have [N1 [N2 [N3 [N4 N5]]]] : x # Θ11 /\ x # u /\ x # Θ12 /\ x # Θ2 /\ x # ts
      by move: (Linear x); rewrite !countE; lia.
    rewrite !substE !substcl_nop // subst_nop // substl_nop //.
    rewrite -!app_assoc -Permutation_middle /=.
    eexists; [done | eexists; [| done]].
    constructor; apply: (ll_Contextual Θ11 _ _ (Θ12 ++ Θ2) (communication _ _ _)).
  - move=> /Exists_exists [C' [/(elem_of_list_split _ _) [Θ21 [Θ22 ->]]
                               /var_is_side_alt [u ->]]] Linear.
    have [N1 [N2 [N3 [N4 N5]]]] : x # Θ1 /\ x # Θ21 /\ x # u /\ x # Θ22 /\ x # ts
      by move: (Linear x); rewrite !countE; lia.
    rewrite !substE !substcl_nop // subst_nop // substl_nop //.
    rewrite Permutation_middle !app_assoc.
    eexists; [done | eexists; [| done]].
    constructor; apply: (ll_Contextual (Θ1 ++ Θ21) _ _ Θ22 (communication _ _ _)).
  - move: (InBody) => /elem_of_count InBody' Linear.
    have [N1 N2] : x # Θ1 /\ x # Θ2 by move: (Linear x); rewrite !countE; lia.
    rewrite !substE !substcl_nop // -Permutation_middle [Coeq _ _]coeq_swap /=.
    eexists; [done | eexists; [| done]].
      by constructor.
Qed.

Definition rstep : relation pexp :=
  reactive_transmission_step ∪ᵣ value_reaction_step.
Definition rsteps : relation pexp := rtc rstep.

Lemma tstep_rstep :
    tstep ≡ᵣ rstep ∪ᵣ unreactive_transmission_step.
Proof. firstorder. Qed.

Lemma rstep_ok : forall {p q}, linear p -> rstep p q -> step p q.
Proof.
  move=> p q Linear [/(reactive_transmission_ok Linear) | /value_reaction_ok] //.
Qed.
Lemma rsteps_ok : forall {p q}, linear p -> rsteps p q -> steps p q.
Proof.
  move=> p q Linear RSteps.
  elim: p q / RSteps Linear; first by done.
  move=> ? ? ? RStep _ IH Linear.
  move: (rstep_ok Linear RStep) => Step.
  etransitivity.
  - right; by constructor.
  - by apply/IH/linear_step.
Qed.

Lemma equiv_rstep_subcommute : subcommute (≡) rstep.
Proof.
  move=> p r [q E [R | V]].
  - move: (E >>> R) => /equiv_reactive_transmission_step_subcommute [q' R' E'].
    exact (rel_inl _ R' >>> E').
  - move: (E >>> V) => /equiv_value_reaction_step_subcommute [q' V' E'].
    exact (rel_inr _ V' >>> E').
Qed.

(** When drawing out a diagram displaying some steps, we'll think of steps that are part of
    [rstep] as moving downward and steps falling under [unreactive_transmission_step] as
    to the right. This will let us think of our important lemmas rather pleasantly at a
    visual level. We'll label an edge u_x for an unreactive transmission at x, r_x for a
    reactive one, and v for a value step; we'll omit subscripts on the transitions if they
    are existentially quantified. *)


Theorem basic_step_rstep : forall {p q}, linear p -> basic_step p q -> rstep p q.
Proof.
  move=> p q Linear BasicStep.
  case: p q / BasicStep Linear => [Θ Θ' ts CReaction | x t Θ ts Reactive] Linear.
  - case: Θ Θ' / CReaction Linear => Θ1 _ _ Θ2 [t x u | | t u t' u'] Linear.
    + left; eexists; apply: rtransmission_r.
      * by left; apply/Exists_app; right; apply/Exists_cons; left; left.
      * have [? [? [? ?]]] : x # Θ1 /\ x # u /\ x # Θ2 /\ x # ts
          by move: (Linear x); rewrite !countE; lia.
        rewrite !substE // !substcl_nop // subst_nop // substl_nop //.
    + by repeat constructor.
    + by repeat constructor.
  - left; eexists; apply: (@rtransmission_l _ []).
    + by right.
    + rewrite /= substcl_nop //.
      move: Reactive (Linear x) => /elem_of_count.
      rewrite !countE; lia.
Qed.

Corollary step_rstep : forall {p q}, linear p -> step p q -> exists2 q', q' ≡ q & rstep p q'.
Proof.
  move=> p q Linear [p_ E [_q Basic E']].
  setoid_rewrite <- E'; clear E'.
  move: (E) (Linear) => {1}-> /basic_step_rstep /(_ Basic) RStep.
  move: (E >>> RStep) => /equiv_rstep_subcommute []; eauto.
Qed.


Definition tsteps : relation pexp := rtc tstep.
Definition usteps : relation pexp := rtc unreactive_transmission_step.

Definition weakly_acyclic1 (C : coeq) : Prop :=
  let '(Coeq t u) := C in
  (if t is var x then x # u else True) /\
  (if u is var y then y # t else True).
Definition weakly_acyclic (p : pexp) : Prop := Forall weakly_acyclic1 (coeqs p).
Instance: Proper ((≡) ==> iff) weakly_acyclic1.
Proof.
  move=> [t u] _ [] //=.
  case: t; case: u; tauto.
Qed.
Instance: Proper ((≡) ==> iff) weakly_acyclic.
Proof.
  apply proper_sym_impl_iff; first by typeclasses eauto.
  move=> [Θ ?] [Θ'' ?] [/soln_equiv_alt_alt /= [Θ' E1 E2] _].
  rewrite /weakly_acyclic /= E1 => /Forall_forall All.
  apply/Forall2_Forall_r/Forall_forall=> // C /All ? ? <- //.
Qed.

(** Counterexample to the statement without the acyclicity hypothesis:
<<
unreactive_transmission_step (x ⊥ y ⊗ x, y ⊥ *; ) (y ⊥ *; )
>>
*)
Theorem linear_unreactive_transmission_step : forall {p q},
    unreactive_transmission_step p q -> weakly_acyclic p -> linear p -> linear q.
Proof.
  move=> _ _ [x [_ t _ ts [Θ1 _ _ Θ2 [_ ->]]
    /unreactive_alt /= [NSide /not_elem_of_count NotIn_ts]]].
  rewrite -Permutation_middle => /Forall_cons [[WA0 NotIn_t] WA].
  move=> Linear y.
  rewrite /= countE count_substcl count_substl !countE.
  case: (decide _) => [-> | NE_xy] /=;
    first by rewrite NotIn_t; left.
  rewrite NotIn_ts Nat.mul_0_r Nat.add_0_r.
  have -> : count x Θ1 + count x Θ2 = 1 by move: (Linear x); rewrite !countE; lia.
  move: (Linear y); rewrite !countE decide_False //; lia.
Qed.
Corollary linear_tstep : forall {p q},
    tstep p q -> weakly_acyclic p -> linear p -> linear q.
Proof.
  move=> p q /tstep_rstep [RStep | /linear_unreactive_transmission_step //] _ Linear.
  apply/linear_step=> //.
    by apply/rstep_ok.
Qed.

Lemma f_neq : forall {A B} (f : A -> B) {x y : A}, f x <> f y -> x <> y.
Proof. firstorder. Qed.

(* TODO refactor this to be less repetitive *)
Lemma lift_value_reaction : forall {t x C},
    var_isn't_side x C -> forall {Θ},
      value_reaction (subst t x C) Θ -> exists2 _Θ, Θ = subst t x _Θ & value_reaction C _Θ.
Proof.
  move=> t x [t' u] [Side1 Side2] Θ.
  move E: (subst _ _ _) => C' ValueReaction.
  case: C' Θ / ValueReaction E; rewrite substE.
  - move=> [].
    case: t' Side1 => //= [x' /(f_neq var) NE | _ _]; first by rewrite subst_var_ne.
    case: u Side2 => //= [x' /(f_neq var) NE | _ _]; first by rewrite subst_var_ne.
      by repeat econstructor.
  - move=> [].
    case: t' Side1 => //= [x' /(f_neq var) NE | _ _]; first by rewrite subst_var_ne.
    case: u Side2 => //= [x' /(f_neq var) NE | _ _]; first by rewrite subst_var_ne.
      by repeat econstructor.
  - move=> t'1 t'2 u1 u2 [].
    case: t' Side1 => //= [x' /(f_neq var) NE | t'1' t'2' _ [<- <-]];
                       first by rewrite subst_var_ne.
    case: u Side2 => //= [x' /(f_neq var) NE | u1' u2' _ [<- <-]];
                      first by rewrite subst_var_ne.
      by repeat econstructor.
  - move=> t'1 t'2 u1 u2 [].
    case: t' Side1 => //= [x' /(f_neq var) NE | t'1' t'2' _ [<- <-]];
                       first by rewrite subst_var_ne.
    case: u Side2 => //= [x' /(f_neq var) NE | u1' u2' _ [<- <-]];
                      first by rewrite subst_var_ne.
      by repeat econstructor.
Qed.

(* TODO need to write some inversion lemmas up above so I don't need to use the inversion
   tactic down here *)
Theorem uv_bend : forall {x},
    unreactive_transmission x >> value_reaction_step ⊆ᵣ
                            value_reaction_step >> transmission x.
(** The resulting [transmission] could be part of rstep or not, depending on
    whether it's a reactive transmission or not. So, our premise is

<<
            u_x
        p -------> q
                   |
                   | v
                   |
                   v
                   r
>>

    and our conclusion is one of

<<
        p
        |
        | v
        |
        v
        q' -------> r
             u_x


        p
        |
        | v
        |
        v
        q'
        |
        | r_x
        |
        v
        r
>>
*)
Proof.
  move=> x p r [q [Θ t Θ'_u ts Del
                    /unreactive_alt [NSide /not_elem_of_count NotIn] VStep]].
  inversion_clear VStep as [? Θ'' ? CValueReaction].
  case: Θ Θ'_u / Del NSide CValueReaction => Θ1 _ _ Θ2 [C_u E] /= /Forall_app [NSide1 NSide2].
  rewrite substE => /el_app1 [[Θ'_u1 -> ] | [Θ'_u2 -> ]].
  - move=> /(el_map1_bwd (Forall_impl _ _ _ NSide1 (@lift_value_reaction _ _)))
           [Θ'_v1 -> CStep_v].
    eexists.
    + constructor; by apply: app_el1.
    + by apply: do_transmission.
  - move=> /(el_map1_bwd (Forall_impl _ _ _ NSide2 (@lift_value_reaction _ _)))
           [Θ'_v2 -> CStep_v].
    eexists.
    + constructor; by apply/app_el2/cons_el2.
    + by apply: do_transmission.
Qed.

Lemma var_is_side_substc : forall {t x y C},
    var_is_side y (subst t x C) <-> var_is_side x C /\ t = var y \/ var_is_side y C /\ x <> y.
Proof.
  move=> t x y [t' u] /=.
  rewrite !subst_is_var; tauto.
Qed.
Lemma reactive_substcl : forall {y t x Θ ts},
    reactive y (subst t x Θ) ts ->
    [\/ reactive x Θ ts /\ t = var y, Exists (var_is_side y) Θ /\ x <> y | y ∈ vars ts].
Proof.
  move=> y t x Θ ts [|]; last by constructor.
  move=> /Exists_fmap /Exists_exists [C [In /var_is_side_substc [] [Side ?]]].
  - constructor 1; split=> //.
    left; apply/Exists_exists; eauto.
  - constructor 2; split=> //.
    apply/Exists_exists; eauto.
Qed.

Lemma lift_del1 : forall {C Θ t x},
    Forall (fun C' => forall ε, del1 C (subst t x C') ε →
                        exists2 _ε, ε = subst t x _ε &
                                    exists2 _C, subst t x _C = C & del1 _C C' _ε) Θ.
Proof.
  move=> [t' u] Θ t x.
  apply/Forall_forall=> -[? ?] _ ε.
  inversion_clear 1 as [? []] => /=; exists []=> //; eexists (Coeq _ _)=> //.
  repeat constructor.
Qed.

Lemma reactive_mult : forall {x Θ ts},
    reactive x Θ ts -> count x Θ + count x ts > 0.
Proof.
  move=> x Θ ts [|].
  - move=> /Exists_exists [C [/(elem_mult (x:=x)) LE /var_is_side_mult GT]]; lia.
  - move=> /elem_of_count; lia.
Qed.

(* TODO refactor this to be less repetitive *)
Theorem ur_bend : forall {p q r x y},
    linear p -> linear q ->
    unreactive_transmission x p q -> reactive_transmission y q r ->
    x <> y /\ exists q', reactive_transmission y p q' /\ transmission x q' r.
(** We go from

<<
            u_x
        p -------> q
                   |
                   | r_y
                   |
                   v
                   r
>>

    to

<<
        p
        |
        | r_y
        |
        v
        q' -------> r
             u_x
>>

    or

<<
        p
        |
        | r_y
        |
        v
        q'
        |
        | r_x
        |
        v
        r
>>

    The latter case is possible in, say:

<<
      p  is   * ⊥ x, * ⊗ x ⊥ y; y
      q  is   * ⊗ * ⊥ y; y
      r  is   ; * ⊗ *
      q' is   * ⊥ x; * ⊗ x
>>
*)
Proof.
  move=> p q r x y Linear1 Linear2 U R.
  move: U R Linear1 Linear2 => [Θ t Θ'_u ts Del
                        /unreactive_alt [NSide /not_elem_of_count NotInBody]].
  case: Θ Θ'_u / Del NSide => Θ1 _ _ Θ2 [C_u E_C_u] /= NSide R Linear1 Linear2.
  inversion_clear R as [? u Θ'' ? Del' Reactive].
  move: Del'; rewrite substE => /el_app1 [[Θ'_u1 ?] | [Θ'_u2 ?]]; subst.
  - move=> D.
    move: D Reactive NSide Linear1 Linear2 => /(el_map1_bwd lift_del1)
      [Θ1m -> [Θ11 _ ε Θ12 [[_u _y] [<- E_y] [C_r E_C_r]]]] Reactive NSide Linear1 Linear2.
    move: NSide; rewrite -!app_assoc [Forall _ _]/= -Permutation_middle =>
      /Forall_cons [NSide_r NSide].
    move: (E_C_r) NSide_r => {1}-> [/nesym NE_xu /nesym NE_xy].
    move: E_y => /subst_is_var [[] | [? _]] //; subst.
    move/(f_neq var) in NE_xy.
    split=> //; eexists; split.
    + rewrite -sim_substl_app -!app_assoc substl_nop //= in Reactive |- *.
      constructor=> //=.
      case: Reactive => [Ex | In]; [left | right]; last by done.
      rewrite -!Permutation_middle; apply Exists_cons_tl.
      move: Ex; rewrite !Exists_exists =>
        -[C [/elem_of_list_fmap [_C [-> In]] /var_is_side_substc [[Side E] | [Side _]]]].
      * move: NSide => /Forall_forall /(_ _ In) /var_isn't_side_alt //.
      * eauto.
    + rewrite /= app_assoc 2!substE.
      evar (q' : pexp).
      rewrite [q' in transmission _ _ q'](_ : _ = ?q'). 2: {
        apply: do_transmission.
        rewrite E_C_u !substE decide_False //.
      }
      have NotIn_y : y # t. {
        apply/not_elem_of_count => /elem_of_count In_y_t.
        rewrite -> E_C_u, E_C_r in Linear1; rewrite -> E_C_r in Linear2.
        rewrite -sim_substl_app in Reactive; rewrite !substE /= decide_False // in Linear2.
        rewrite substl_nop //= -!app_assoc /= in Reactive Linear2.
        case: (reactive_substcl Reactive) => [[] | |].
        - move=> /(fun x (k : ~_) => k x); rewrite unreactive_alt; case.
          split; [done | by apply/not_elem_of_count].
        - move=> [/Exists_exists [C [/(elem_mult (x:=y)) LE /var_is_side_mult GT]] _].
          move: LE GT (Linear1 y).
          rewrite !countE; lia.
        - move=> /elem_of_count In_y_ts.
          move: (Linear1 y).
          rewrite !countE; lia.
      }
      rewrite -!sim_substl_app substE /sim_subst /list_subst -!list_fmap_compose.
      f_equal; apply/list_fmap_ext=> //.
      * move=> C /=.
        rewrite -![term_subst _ _]/(subst _ _ _) -[L in L = _]substc_substc2 // subst_nop //.
      * move=> t0 /=.
        rewrite -![term_subst _ _]/(subst _ _ _) -[L in L = _]subst_subst2 //
                [subst _ y t]subst_nop //.
  - move=> D.
    move: D Reactive NSide Linear1 Linear2 => /(el_map1_bwd lift_del1)
      [Θ2m -> [Θ21 _ ε Θ22 [[_u _y] [<- E_y] [C_r E_C_r]]]] Reactive NSide Linear1 Linear2.
    move: NSide; rewrite [Forall _ _]/= -![in Forall _ _]Permutation_middle =>
      /Forall_cons [NSide_r NSide].
    move: (E_C_r) NSide_r => {1}-> [/nesym NE_xu /nesym NE_xy].
    move: E_y => /subst_is_var [[] | [? _]] //; subst.
    move/(f_neq var) in NE_xy.
    split=> //; eexists; split.
    + rewrite -sim_substl_app substl_nop //= in Reactive.
      rewrite cons_middle 2!app_assoc -[_ ++ Θ21]app_assoc.
      constructor=> //=.
      case: Reactive => [Ex | In]; [left | right]; last by done.
      rewrite -!app_assoc /= -Permutation_middle; apply Exists_cons_tl.
      move: Ex; rewrite !Exists_exists =>
        -[C [/elem_of_list_fmap [_C [-> In]] /var_is_side_substc [[Side E] | [Side _]]]].
      * move: NSide => /Forall_forall /(_ _ In) /var_isn't_side_alt //.
      * eauto.
    + rewrite /= -app_assoc /= 2!substE.
      evar (q' : pexp).
      rewrite [q' in transmission _ _ q'](_ : _ = ?q'). 2: {
        apply: do_transmission.
        rewrite E_C_u !substE decide_False //.
      }
      have NotIn_y : y # t. {
        apply/not_elem_of_count => /elem_of_count In_y_t.
        rewrite -> E_C_u, E_C_r in Linear1; rewrite -> E_C_r in Linear2.
        rewrite -sim_substl_app in Reactive; rewrite !substE /= decide_False // in Linear2.
        rewrite substl_nop //= /= in Reactive Linear2.
        case: (reactive_substcl Reactive) => [[] | |].
        - move=> /(fun x (k : ~_) => k x); rewrite unreactive_alt; case.
          split; [done | by apply/not_elem_of_count].
        - move=> [/Exists_exists [C [/(elem_mult (x:=y)) LE /var_is_side_mult GT]] _].
          move: LE GT (Linear1 y).
          rewrite !countE; lia.
        - move=> /elem_of_count In_y_ts.
          move: (Linear1 y).
          rewrite !countE; lia.
      }
      rewrite -!sim_substl_app substE /sim_subst /list_subst -!list_fmap_compose.
      f_equal; apply/list_fmap_ext=> //.
      * move=> C /=.
        rewrite -![term_subst _ _]/(subst _ _ _) -[L in L = _]substc_substc2 // subst_nop //.
      * move=> t0 /=.
        rewrite -![term_subst _ _]/(subst _ _ _) -[L in L = _]subst_subst2 //
                [subst _ y t]subst_nop //.
Qed.

Corollary u_bend : forall {p q r x},
    linear p -> linear q ->
    unreactive_transmission x p q -> rstep q r ->
    exists q', rstep p q' /\ transmission x q' r.
Proof.
  move=> p q r x Linear1 Linear2 U [[y R] | V].
  - move: (ur_bend Linear1 Linear2 U R) => [_ [q' [R' T]]].
    eexists; split; [left; eauto | done].
  - move: (U >>> V) => /uv_bend [q' V' U'].
    eexists; split; [by right | done].
Qed.

Definition corr : relation pexp := flip usteps.

Section UsingAcyclicity.
  Variable acyclic : pexp -> Prop.
  Context `{!Proper ((≡) ==> iff) acyclic}.
  Hypothesis acyclicE : forall {p}, acyclic p -> weakly_acyclic p.
  Hypothesis acyclic_preserved : forall {p q}, tstep p q -> acyclic p -> acyclic q.

(* One of the main goals. *)
  Theorem defer_unreactives : forall {p r},
      linear p -> acyclic p -> tsteps p r ->
      exists q, rsteps p q /\ usteps q r.
(** Concept: a [tsteps] is a zigzag path down and to the right. This theorem states that
    we can reshape this into a big L with the same start and end points. We can do this by
    induction on the path. The base case is trivial. In the step case, we first reshape the
    remainder of the path after the first step. If the first step is then a step down, we
    can just prepend it. Otherwise, we have work to do. Things look like this:

<<
        p -------> p'
                   |
                   |
                   |
                   v
                   .
                   .
                   .
                   |
                   |
                   |
                   v
                   q -------> ... -------> r
>>

    What we can do is iteratively walk the horizontal part downward using our prior lemmas
    about bends. At each step, we have something like

<<
        p
        |
        |
        |
        v
        .
        .
        .
        |
        |
        |
        v
        a -------> b
                   |
                   |
                   |
                   v
                   .
                   .
                   .
                   |
                   |
                   |
                   v
                   q -------> ... -------> r
>>

    If the second vertical section is in fact of height zero---that is, if b is actually
    q---then we are done; we have an L. Otherwise, we have some c below b, and a bend:

<<
        a -------> b
                   |
                   |
                   |
                   v
                   c
>>

    We can use our lemmas to turn this bend into one of

<<
        a
        |
        |
        |
        v
        b' -------> c


        a
        |
        |
        |
        v
        b'
        |
        |
        |
        v
        c
>>

    In the former case, we have decreased the height of the lower vertical segment, so we
    can continue downward. In the latter case, we're done entirely, because we now have
    an L!
*)
  Proof.
    move=> p r Linear Acyclic TSteps.
    elim: p r / TSteps Linear Acyclic; first by repeat econstructor.
    move=> p p' r TStep _ IH Linear Acyclic.
    move: (acyclic_preserved _ _ TStep Acyclic)
      (linear_tstep TStep (acyclicE _ Acyclic) Linear) => Acyclic' Linear'.
    move: (IH Linear' Acyclic') TStep => [q [RSteps USteps]] /tstep_rstep [RStep | UStep];
      first by (exists q; split; [econstructor |]).
    elim: p' q / RSteps p Linear Acyclic UStep USteps {IH Linear' Acyclic'};
      first by move=> q p *; exists p; split; by econstructor.
    move=> b c q RStep RSteps IH p Linear Acyclic UStep USteps.
    move: UStep (rel_inr rstep UStep) => [x Unreactive] /tstep_rstep TStep.
    move: (linear_tstep TStep (acyclicE _ Acyclic) Linear) => Linear'.
    move: (u_bend Linear Linear' Unreactive RStep) => [b' [RStep' [R | U]]].
    - exists q; split; last by done.
      econstructor; first by done.
      econstructor; last by done.
      left; by eexists.
    - move: {U} (ex_intro _ _ U : unreactive_transmission_step _ _) => UStep'.
      move: (rel_inl unreactive_transmission_step RStep') => /tstep_rstep TStep'.
      move: (acyclic_preserved _ _ TStep' Acyclic)
        (linear_tstep TStep' (acyclicE _ Acyclic) Linear) => {Linear'} Acyclic' Linear'.
      move: (IH _ Linear' Acyclic' UStep' USteps) => [q' [Rsteps' USteps']].
      exists q'; split; last by done.
        by econstructor.
  Qed.

  Lemma refute_unreactive : forall {p q},
      linear p -> weakly_acyclic p -> cut_free q -> ~unreactive_transmission_step p q.
  Proof.
    move=> p q Linear WA CF [x Transmission].
    case: p q / Transmission Linear WA CF => Θ t Θ' ts Del /=.
    case: Θ Θ' / Del => Θ1 _ _ Θ2 [_ ->] /=
      /unreactive_alt [_ /not_elem_of_count NotIn_ts] /(_ x) Linear_x
      /Forall_middle /= [_ NotIn_t]
      /map_eq_nil /app_nil [? ?]; subst.
    move: Linear_x; rewrite /= !countE NotIn_ts => L.
    lia.
  Qed.
  Theorem clear_unreactives : forall {p q},
      linear p -> acyclic p -> cut_free q -> usteps p q -> p = q.
  Proof.
    move=> p q Linear Acyclic CF /(rtc_inv_r _ _) [// | [p' [USteps UStep]]].
    have [Linear' /acyclicE WA] : linear p' /\ acyclic p'. {
      elim: p p' / USteps Linear Acyclic {UStep} => // p p' p'' UStep _ IH Linear Acyclic.
      apply: IH.
      - apply: linear_tstep.
        + by apply/tstep_rstep/rel_inr.
        + by apply: acyclicE.
        + done.
      - apply: acyclic_preserved.
        + by apply/tstep_rstep/rel_inr.
        + done.
    }
    case: (refute_unreactive _ _ _ UStep) => //.
  Qed.

  (* TODO factor out repeated preservation subproof *)
  Theorem eval : forall {p q}, linear p -> acyclic p -> cut_free q -> tsteps p q -> steps p q.
  Proof.
    move=> p q Linear Acyclic CF TSteps.
    move: (defer_unreactives Linear Acyclic TSteps) => [q0 [RSteps USteps]].
    have [Linear' Acyclic'] : linear q0 /\ acyclic q0. {
      elim: p q0 / RSteps Linear Acyclic {TSteps USteps} =>
        // p p' q0 RStep _ IH Linear Acyclic.
      apply: IH.
      - apply: linear_tstep.
        + by apply/tstep_rstep/rel_inl.
        + by apply: acyclicE.
        + done.
      - apply: acyclic_preserved.
        + by apply/tstep_rstep/rel_inl.
        + done.
    }
    move: (clear_unreactives Linear' Acyclic' CF USteps) => ?; subst.
      by apply/rsteps_ok.
  Qed.
End UsingAcyclicity.
