{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE FlexibleContexts #-}
module PE2 where

import Data.Map (Map)
import qualified Data.Map as M
import qualified Data.Set as S
import Data.List
import Data.String
import Data.Functor.Identity
import Control.Monad.State
import Control.Applicative

data Name = Name String String deriving (Show, Eq, Ord)
instance IsString Name where fromString s = Name s ""

data Term
  = Var Name
  | Unit
  | Counit
  | Times Term Term
  | Par   Term Term
  | Inl Term
  | Inr Term
  | With [Name] PrfExp PrfExp
  | Der Term
  | Weak
  | Contr Term Term
  | Ofc [Name] PrfExp
  deriving (Show)
-- TODO use snoc lists or something, there are 'init's and 'last's in this code
data PrfExp = PrfExp [Coeq] [Term] deriving (Show)
data Coeq = Cut Term Term deriving (Show)

var :: String -> Term
var = Var . fromString
instance IsString Term where fromString = var

infixr 5 ⊗
infixr 4 ⅋
infixr 7 @@
(⊗), (⅋), (@@) :: Term -> Term -> Term
(⊗) = Times
(⅋) = Par
(@@) = Contr

infix 3 ⊥
(⊥) :: Term -> Term -> Coeq
(⊥) = Cut

switch :: Coeq -> Coeq
switch (Cut t1 t2) = Cut t2 t1

class TraverseNames a where
  traverseNames :: Applicative f => (Name -> f Name) -> a -> f a

mapNames :: TraverseNames a => (Name -> Name) -> a -> a
mapNames f = runIdentity . traverseNames (Identity . f)

instance TraverseNames Name where
  traverseNames = id

instance TraverseNames Term where
  traverseNames f t' = case t' of
    Var n  -> Var <$> f n
    Unit   -> pure Unit
    Counit -> pure Counit
    Times t1 t2 -> liftA2 Times (traverseNames f t1) (traverseNames f t2)
    Par   t1 t2 -> liftA2 Par (traverseNames f t1) (traverseNames f t2)
    Inl t -> fmap Inl (traverseNames f t)
    Inr t -> fmap Inr (traverseNames f t)
    With ns e1 e2 ->
      liftA3 With (traverse f ns) (traverseNames f e1) (traverseNames f e2)
    Der t -> fmap Der (traverseNames f t)
    Weak -> pure Weak
    Contr t1 t2 -> liftA2 Contr (traverseNames f t1) (traverseNames f t2)
    Ofc ns e -> liftA2 Ofc (traverse f ns) (traverseNames f e)

instance TraverseNames PrfExp where
  traverseNames f (PrfExp coeqs terms) = liftA2 PrfExp
    (traverseNames f coeqs) (traverseNames f terms)

instance TraverseNames Coeq where
  traverseNames f (Cut t1 t2) =
    liftA2 Cut (traverseNames f t1) (traverseNames f t2)

instance TraverseNames a => TraverseNames [a] where
  traverseNames = traverse . traverseNames

renameL, renameR :: TraverseNames a => a -> a
renameL = mapNames (\(Name x0 lr) -> Name x0 (lr ++ "l"))
renameR = mapNames (\(Name x0 lr) -> Name x0 (lr ++ "r"))

ppName :: Name -> String
ppName (Name x0 lr) = x0 ++ lr

ppTermP :: Int -> Term -> String
ppTermP prec t' = case t' of
  Var n  -> ppName n
  Unit   -> "*"
  Counit -> "(*)"
  Times t1 t2 -> p 5 $ ppTermP 6 t1 ++ " ⊗ " ++ ppTermP 5 t2
  Par   t1 t2 -> p 4 $ ppTermP 5 t1 ++ " ⅋ " ++ ppTermP 4 t2
  Inl t -> "inl " ++ ppTermP 10 t
  Inr t -> "inr " ++ ppTermP 10 t
  With ns e1 e2 -> p 9 $ concat [
    unwords (map ppName ns), "(", ppPrfExp e1, " & ", ppPrfExp e2, ")"]
  Der t -> "?" ++ ppTermP 10 t
  Weak -> "_"
  Contr t1 t2 -> p 7 $ ppTermP 8 t1 ++ " @ " ++ ppTermP 7 t2
  Ofc ns e -> p 9 $ concat [unwords (map ppName ns), "(", ppPrfExp e, ")"]
  where p prec' s | prec <= prec' = s
                  | otherwise = "(" ++ s ++ ")"

ppTerm :: Term -> String
ppTerm = ppTermP (-1)

ppPrfExp :: PrfExp -> String
ppPrfExp (PrfExp coeqs terms) =
  intercalate ", " (map ppCoeq coeqs) ++ (if null terms then ";" else "; ") ++
  intercalate ", " (map ppTerm terms)

ppCoeq :: Coeq -> String
ppCoeq (Cut t1 t2) = ppTerm t1 ++ " ⊥ " ++ ppTerm t2


type Substitution = Map Name Term
type StepM = StateT Substitution Maybe

processCoeq :: Coeq -> StepM [Coeq]
processCoeq coeq = go coeq <|> map switch <$> go (switch coeq)
  where
    go :: Coeq -> StepM [Coeq]
    go coeq' = case coeq' of
      Cut (Var n) t -> do
        nSubst <- gets (M.lookup n)
        case nSubst of
          Nothing -> modify (M.insert n t) >> return []
          Just t' -> modify (M.delete n) >> return [t' ⊥ t]
      Cut Unit Counit -> return []
      Cut (Times t1 t2) (Par t1' t2') -> return [t1 ⊥ t1', t2 ⊥ t2']
      Cut (With ns (PrfExp cs ts) _) (Inl t) | not (null ts) ->
        return (cs ++ zipWith Cut (map Var ns) (init ts) ++ [last ts ⊥ t])
      Cut (With ns _ (PrfExp cs ts)) (Inr t) | not (null ts) ->
        return (cs ++ zipWith Cut (map Var ns) (init ts) ++ [last ts ⊥ t])
      Cut (Ofc ns (PrfExp cs ts)) (Der t) | not (null ts) ->
        return (cs ++ zipWith Cut (map Var ns) (init ts) ++ [last ts ⊥ t])
      Cut (Ofc ns _) Weak -> return (zipWith Cut (map Var ns) (repeat Weak))
      Cut t@(Ofc ns _) (Contr t1 t2) ->
        let contractName n = Var n ⊥ renameL (Var n) @@ renameR (Var n)
        in return (map contractName ns ++ [renameL t ⊥ t1, renameR t ⊥ t2])
      _ -> lift Nothing

termSubst :: Term -> StepM Term
termSubst t' = case t' of
  Var n -> do
    nSubst <- gets (M.lookup n)
    case nSubst of
      Nothing -> return (Var n)
      Just t -> modify (M.delete n) >> termSubst t
  Unit   -> return Unit
  Counit -> return Counit
  Times t1 t2 -> liftA2 Times (termSubst t1) (termSubst t2)
  Par   t1 t2 -> liftA2 Par (termSubst t1) (termSubst t2)
  Inl t -> fmap Inl (termSubst t)
  Inr t -> fmap Inr (termSubst t)
  With ns e1 e2 -> liftA2 (With ns) (prfExpSubst e1) (prfExpSubst e2)
  Der t -> fmap Der (termSubst t)
  Weak -> return Weak
  Contr t1 t2 -> liftA2 Contr (termSubst t1) (termSubst t2)
  Ofc ns e -> fmap (Ofc ns) (prfExpSubst e)

prfExpSubst :: PrfExp -> StepM PrfExp
prfExpSubst (PrfExp coeqs terms) =
  liftA2 PrfExp (traverse coeqSubst coeqs) (traverse termSubst terms)

coeqSubst :: Coeq -> StepM Coeq
coeqSubst (Cut t1 t2) = liftA2 Cut (termSubst t1) (termSubst t2)


step :: PrfExp -> StepM (Either [Term] PrfExp)
step (PrfExp [] ts) = return (Left ts)
step (PrfExp (c:cs) ts) = do
  new <- processCoeq c
  return (Right (PrfExp (new ++ cs) ts))

steps :: PrfExp -> StepM PrfExp
steps e = do
  stepped <- step e
  case stepped of
    Left ts -> do
      ts' <- traverse termSubst ts
      subst <- get
      let substCs = map (\(n, t) -> Var n ⊥ t) (M.toList subst)
      return (PrfExp substCs ts')
    Right e' -> steps e'

steps' :: PrfExp -> StepM [PrfExp]
steps' e = do
  stepped <- step e
  case stepped of
    Left ts -> do
      ts' <- traverse termSubst ts
      subst <- get
      let substCs = map (\(n, t) -> Var n ⊥ t) (M.toList subst)
      return [PrfExp substCs ts']
    Right (PrfExp cs ts) -> do
      ts' <- traverse termSubst ts
      subst <- get
      let substCs = map (\(n, t) -> Var n ⊥ t) (M.toList subst)
          logged = PrfExp (substCs ++ cs) ts'
      (logged:) <$> steps' (PrfExp cs ts')

eval :: PrfExp -> Maybe PrfExp
eval e = evalStateT (steps e) M.empty


type ToCM = State (Int, Map String String)

mkFun :: String -> ToCM String
mkFun body = do
  (n, m) <- get
  let name = 'v':show n
  put (n + 1, M.insert name body m)
  return name

traverseTopNamesT :: Applicative f => (Name -> f Name) -> Term -> f Term
traverseTopNamesT f t' = case t' of
  Var n  -> Var <$> f n
  Unit   -> pure Unit
  Counit -> pure Counit
  Times t1 t2 -> liftA2 Times (traverseTopNamesT f t1) (traverseTopNamesT f t2)
  Par   t1 t2 -> liftA2 Par (traverseTopNamesT f t1) (traverseTopNamesT f t2)
  Inl t -> fmap Inl (traverseTopNamesT f t)
  Inr t -> fmap Inr (traverseTopNamesT f t)
  With ns e1 e2 -> liftA3 With (traverse f ns) (pure e1) (pure e2)
  Der t -> fmap Der (traverseTopNamesT f t)
  Weak -> pure Weak
  Contr t1 t2 -> liftA2 Contr (traverseTopNamesT f t1) (traverseTopNamesT f t2)
  Ofc ns e -> liftA2 Ofc (traverse f ns) (pure e)

traverseTopNamesE :: Applicative f => (Name -> f Name) -> PrfExp -> f PrfExp
traverseTopNamesE f (PrfExp coeqs terms) = liftA2 PrfExp
  (traverse (traverseTopNamesC f) coeqs) (traverse (traverseTopNamesT f) terms)

traverseTopNamesC :: Applicative f => (Name -> f Name) -> Coeq -> f Coeq
traverseTopNamesC f (Cut t1 t2) =
  liftA2 Cut (traverseTopNamesT f t1) (traverseTopNamesT f t2)

termToC :: Term -> ToCM String -- output an expression
termToC t' = case t' of
  Var n  -> pure (ppName n) -- TODO we can't /actually/ rely on this...
  Unit   -> pure "unit()"
  Counit -> pure "counit()"
  Times t1 t2 -> do
    s1 <- termToC t1; s2 <- termToC t2
    return . concat $ ["tens(", s1, ", ", s2, ")"]
  Par   t1 t2 -> do
    s1 <- termToC t1; s2 <- termToC t2
    return . concat $ ["par(", s1, ", ", s2, ")"]
  Inl t -> do s <- termToC t; return . concat $ ["inl(", s, ")"]
  Inr t -> do s <- termToC t; return . concat $ ["inr(", s, ")"]
  With ns e1 e2 -> do
    let len = show (length ns)
        clos = if null ns then "NULL" else concat ["copy_nodes(", len,
          ", (Node[]){", intercalate ", " (map ppName ns), "})"]
    f1 <- prfExpToC e1 >>= mkFun; f2 <- prfExpToC e2 >>= mkFun
    return . concat $ ["with(", len, ", ", clos, ", ", f1, ", ", f2, ")"]
  Der t -> do s <- termToC t; return . concat $ ["der(", s, ")"]
  Weak -> pure "weak()"
  Contr t1 t2 -> do
    s1 <- termToC t1; s2 <- termToC t2
    return . concat $ ["contr(", s1, ", ", s2, ")"]
  Ofc ns e -> do
    let len = show (length ns)
        clos = if null ns then "NULL" else concat ["copy_nodes(", len,
          ", (Node[]){", intercalate ", " (map ppName ns), "})"]
    f <- prfExpToC e >>= mkFun
    return . concat $ ["ofc(", len, ", ", clos, ", ", f, ")"]

prfExpToC :: PrfExp -> ToCM String -- output a series of statements
prfExpToC e = do
  let rename nm@(Name x0 lr) = do
        seen <- get
        if S.member nm seen then return (Name x0 (lr ++ ['_']))
                            else modify (S.insert nm) >> return nm
      (PrfExp coeqs terms, nms) = runState (traverseTopNamesE rename e) S.empty
      nus = "    " ++ foldMap (\nm -> "NU_(" ++ ppName nm ++ ");") nms
      cutClos i s = "    cut(" ++ s ++ ", clos[" ++ show (i :: Int) ++ "]);"
  cuts1 <- unlines <$> traverse (fmap ("    " ++) . coeqToC) coeqs
  cuts2 <- unlines . zipWith cutClos [0..] <$> traverse termToC (init terms)
  term' <- termToC (last terms)
  return . concat $ [nus, "\n", cuts1, cuts2, "    ", "return ", term', ";\n"]

coeqToC :: Coeq -> ToCM String -- output a statement
coeqToC (Cut t1 t2) = fmap concat . sequence $
  [pure "cut(", termToC t1, pure ", ", termToC t2, pure ");"]

compileToC :: PrfExp -> String
compileToC e = concat [decls, defs, "Node main_(Nodes clos) {\n", main_, "}\n"]
  where (main_, (_, funs)) = runState (prfExpToC e) (0, M.empty)
        decls = concatMap (\name ->
          concat ["Node ", name, "(Nodes clos);\n"]) (M.keys funs)
        defs = concatMap (\(name, body) ->
          concat ["Node ", name, "(Nodes clos) {\n", body, "}\n"])
            (M.assocs funs)


type FreshM = StateT Int Maybe

fresh :: MonadState Int m => m Name
fresh = do
  n <- get
  put (n + 1)
  return (Name ('v':show n) "f")

freshV :: MonadState Int m => m Term
freshV = Var <$> fresh

freshVs :: MonadState Int m => Int -> m [Term]
freshVs n = replicateM n freshV

cf :: Term -> PrfExp
cf = PrfExp [] . pure

cut :: PrfExp -> PrfExp -> PrfExp
cut (PrfExp coeqs1 terms1@(_:_)) (PrfExp coeqs2 terms2@(_:_)) =
  PrfExp (coeqs1 ++ coeqs2 ++ [last terms1 ⊥ last terms2])
         (init terms1 ++ init terms2)
cut _ _ = error "bad cut" -- TODO ehhh

-- TODO double check that all of this stuff is actually well-typed :\

-- ⊢ P : A ⇒ ⊢ ofcMap0 t : !A
ofcMap0 :: PrfExp -> FreshM Term
ofcMap0 p = return $ Ofc [] p

-- ⊢ P : A ⊸ B ⇒ ⊢ ofcMap1 P : !A ⊸ !B
ofcMap1 :: PrfExp -> FreshM Term
ofcMap1 p = do
  [x@(Var xn), a, b] <- freshVs 3
  return $ x ⅋ Ofc [xn] (p `cut` PrfExp [] [Der a, b, a ⊗ b])

-- nat is defined by (∀A.!(A ⊸ A) ⊸ A ⊸ A)

-- ⊢ zero : nat
zero :: FreshM Term
zero = do
  [z] <- freshVs 1
  return $ Weak ⅋ z ⅋ z

church :: Int -> FreshM Term
church 0 = zero
church n = do
  z:vs <- freshVs (n + 1)
  return $ foldr1 (@@) (zipWith (\x y -> Der (x ⊗ y)) (z:vs) vs) ⅋ z ⅋ last vs

-- a bit imprecise, but...
unchurch :: Term -> Maybe Int
unchurch (Par apps (Par _ _)) = go apps
  where go Weak = Just 0
        go (Contr u v) = liftA2 (+) (go u) (go v)
        go (Der (Times _ _)) = Just 1
        go _ = Nothing
unchurch _ = Nothing

-- ⊢ suc : nat ⊸ nat
suc :: FreshM Term
suc = do
  [s, one, z, out] <- freshVs 4
  return $ (s ⊗ one ⊗ out) ⅋ (s @@ Der (z ⊗ one) ⅋ z ⅋ out)

-- ⊢ promoteNat : nat ⊸ !nat
promoteNat :: FreshM Term
promoteNat = do
  [out] <- freshVs 1
  suc' <- ofcMap0 . cf =<< ofcMap1 . cf =<< suc
  zero' <- ofcMap0 . cf =<< zero
  -- lhs of the ⅋ is !(!nat ⊸ !nat) ⊗ !nat ⊗ (!nat)^⊥, hence
  -- ∃A.!(A ⊸ A) ⊗ A ⊗ A^⊥
  return $ (suc' ⊗ zero' ⊗ out) ⅋ out

-- ⊢ add : nat ⊸ nat ⊸ nat
add :: FreshM Term
add = do
  [s0, s1, z, n, out] <- freshVs 5
  return $ (s0 ⊗ n ⊗ out) ⅋ (s1 ⊗ z ⊗ n) ⅋ (s0 @@ s1 ⅋ z ⅋ out)

-- ⊢ mul : nat ⊸ nat ⊸ nat
mul :: FreshM Term
mul = do
  [out, m, m'@(Var m'n), m'', n, n'] <- freshVs 6
  promoteNat' <- promoteNat
  add' <- add
  zero' <- zero
  let addm = Ofc [m'n] (PrfExp [add' ⊥ m'' ⊗ n ⊗ n'] [Der m'', n ⅋ n'])
      Just (PrfExp [] [t]) = eval $ PrfExp [promoteNat' ⊥ m ⊗ m']
        [(addm ⊗ zero' ⊗ out) ⅋ m ⅋ out]
  return t

-- bool is defined by 1 ⊕ 1

-- ⊢ tt : bool    ⊢ ff : bool
tt, ff :: Term
tt = Inl Unit
ff = Inr Unit

-- ⊢ negb : bool ⊸ bool
negb :: FreshM Term
negb = do
  [b@(Var bn)] <- freshVs 1
  return $ With [bn] (PrfExp [] [ff, Counit]) (PrfExp [] [tt, Counit]) ⅋ b

-- int_k is defined by int_0 := 1, int_(S k) := int_k ⊗ bool

-- ⊢ zero_k k : int_k
zero_k :: Int -> FreshM Term
zero_k n = return $ iterate (\t -> Times t ff) Unit !! n

-- ⊢ suc_k k : int_k ⊸ int_k
suc_k :: Int -> FreshM Term
suc_k 0 = return $ Counit ⅋ Unit
suc_k n = do
  [high@(Var highn), out@(Var outn), high1, high1', high2] <- freshVs 5
  r <- suc_k (n - 1)
  return $ (high ⅋ With [highn, outn]
      (PrfExp [r ⊥ high1 ⊗ high1'] [high1, high1' ⊗ ff, Counit])
      (PrfExp [] [high2, high2 ⊗ tt, Counit]))
    ⅋ out

-- ⊢ toFixedWidth k : nat ⊸ int_k
toFixedWidth :: Int -> FreshM Term
toFixedWidth n = do
  [out] <- freshVs 1
  zero_n <- zero_k n
  suc_n <- ofcMap0 . cf =<< suc_k n
  return $ (suc_n ⊗ zero_n ⊗ out) ⅋ out

-- ⊢ add_k k : int_k ⊸ int_k ⊸ int_k
add_k :: Int -> FreshM PrfExp
add_k 0 = return . cf $ Counit ⅋ Counit ⅋ Unit
add_k n = do
  [hil@(Var hiln), hir@(Var hirn), lor@(Var lorn),
   out@(Var outn), out'@(Var out'n),
   hil1@(Var hil1n), hil11, hil12, hil2,
   hir1@(Var hir1n), hir11, hir12, hir2,
   lor', s1, s1', s2, s3,
   rv@(Var rvn), r1@(Var r1n)] <- freshVs 20
  s <- suc_k (n - 1)
  r <- add_k (n - 1)
  return $ r `cut` PrfExp [] [(hil ⅋ With [hiln, hirn, lorn, outn, rvn]
      (PrfExp [] [hil1, hir1, With [hil1n, hir1n, out'n, r1n]
          (PrfExp [s ⊥ s1 ⊗ s1']
                  [hil11, hir11, s1' ⊗ ff, hil11 ⊗ hir11 ⊗ s1, Counit])
          (PrfExp [] [hil12, hir12, s2 ⊗ tt, hil12 ⊗ hir12 ⊗ s2, Counit]),
        out', r1, Counit])
      (PrfExp [] [hil2, hir2, lor', s3 ⊗ lor', hil2 ⊗ hir2 ⊗ s3, Counit]))
    ⅋ (hir ⅋ lor) ⅋ out, rv]

add_k' :: Int -> FreshM PrfExp
add_k' 0 = return . cf $ Counit ⅋ Counit ⅋ Unit
add_k' n = do
  [ah, r@(Var rn), b@(Var bn), o@(Var on),
   b1, b'h1, b'l1, oh1,
   bh2, bl2, oh2] <- freshVs 11
  s <- suc_k (n - 1)
  r_ <- add_k' (n - 1)
  let p = PrfExp [s ⊥ b1 ⊗ (b'h1 ⅋ b'l1)] [b1, b'h1 ⊗ oh1, oh1 ⊗ b'l1, Counit]
      q = PrfExp [] [bh2 ⅋ bl2, bh2 ⊗ oh2, oh2 ⊗ bl2, Counit]
  return $ r_ `cut` PrfExp [] [(ah ⅋ With [bn, rn, on] p q) ⅋ b ⅋ o, ah ⊗ r]


lemCut :: PrfExp
lemCut = [Der (Inl "x") @@ Der (Inr "x") ⊥
          Ofc ["f0'", "f1'", "o'"] ([] `PrfExp`
            ["f0", "f1", Der "o", With ["f0", "f1", "o"]
              ([] `PrfExp` [Der ("x" ⊗ "y"), Weak, "y", "x"])
              ([] `PrfExp` [Weak, Der ("x'" ⊗ "y'"), "y'", "x'"])])]
         `PrfExp` ["f0'", "f1'", "o'"]

addCut :: PrfExp
addCut = [renameL one ⊥ "s1" ⊗ "n" ⊗ "out", renameR two ⊥ "s2" ⊗ "z" ⊗ "n"]
         `PrfExp` ["s1" @@ "s2" ⅋ "z" ⅋ "out"]
  where one = Der ("z" ⊗ "one") ⅋ "z" ⅋ "one"
        two = Der ("z" ⊗ "one") @@ Der ("one" ⊗ "two") ⅋ "z" ⅋ "two"

dupCut :: PrfExp
dupCut = [renameL dup ⊥ renameR two ⊗ "out"] `PrfExp` ["out"]
  where dup = Ofc [] (PrfExp [] [
            (("s0" ⊗ "z0" ⊗ "n0") ⅋ ("s1" ⊗ "z1" ⊗ "n1")) ⅋
            ((("s0" @@ Der ("n0" ⊗ "out0")) ⅋ "z0" ⅋ "out0") ⊗
             (("s1" @@ Der ("n1" ⊗ "out1")) ⅋ "z1" ⅋ "out1"))]) ⊗
          ((Weak ⅋ "z2" ⅋ "z2") ⊗ (Weak ⅋ "z3" ⅋ "z3")) ⊗ "out" ⅋ "out"
        two = Der ("z" ⊗ "one") @@ Der ("one" ⊗ "two") ⅋ "z" ⅋ "two"

mulCut :: PrfExp
Just mulCut = flip evalStateT 0 $ do
  mul' <- ofcMap0 . cf =<< mul
  five' <- ofcMap0 . cf =<< church 5
  [mul1, mul2, mul3, five1, five2, five3, five4, out0, out1, out] <- freshVs 10
  return $ PrfExp [mul' ⊥ Der mul1 @@ Der mul2 @@ Der mul3,
                   five' ⊥ Der five1 @@ Der five2 @@ Der five3 @@ Der five4,
                   mul1 ⊥ five1 ⊗ five2 ⊗ out0, mul2 ⊥ out0 ⊗ five3 ⊗ out1,
                   mul3 ⊥ out1 ⊗ five4 ⊗ out]
                  [out]

toFWCut :: PrfExp
Just toFWCut = flip evalStateT 0 $ do
  fw <- toFixedWidth 10
  num <- church 30
  out <- freshV
  return $ PrfExp [fw ⊥ num ⊗ out] [out]

add_kCut :: PrfExp
Just add_kCut = flip evalStateT 0 $ do
  fw1 <- toFixedWidth 8
  fw2 <- toFixedWidth 8
  num1 <- church 7
  num2 <- church 20
  add8 <- add_k 8
  [out, int1, int2] <- freshVs 3
  return $ add8 `cut` PrfExp [fw1 ⊥ num1 ⊗ int1, fw2 ⊥ num2 ⊗ int2]
                             [out, int1 ⊗ int2 ⊗ out]

add_kCut' :: PrfExp
Just add_kCut' = flip evalStateT 0 $ do
  fw1 <- toFixedWidth 8
  fw2 <- toFixedWidth 8
  num1 <- church 7
  num2 <- church 20
  add8 <- add_k' 8
  [out, int1, int2] <- freshVs 3
  return $ add8 `cut` PrfExp [fw1 ⊥ num1 ⊗ int1, fw2 ⊥ num2 ⊗ int2]
                             [out, int1 ⊗ int2 ⊗ out]


main :: IO ()
main = case eval lemCut of
  Nothing -> putStrLn "wat"
  Just e  -> putStrLn . ppPrfExp $ e

