#include <stdlib.h>
#include <string.h>
#include <stdio.h>

typedef struct Term {
    enum TermClass {
        LINK, UNIT, COUNIT, TENS, PAR, INL, INR, WITH, DER, WEAK, CONTR, OFC,
    } class;
    union TermU {
        /* pointer to the other instance of the same variable */
        struct Term *link;
        /* unit and counit do not require any fields */
        struct {struct Term *l, *r;} tens, par;
        struct Term *inl, *inr;
        /* clos should be length len; l and r should take clos. */
        struct {int len; struct Term **clos;
            struct Term *(*l)(struct Term **), *(*r)(struct Term **);} with;
        /* weakening does not require any field */
        struct Term *der;
        struct {struct Term *l, *r;} contr;
        /* clos should be length len; body should take clos. */
        struct {int len; struct Term **clos;
            struct Term *(*body)(struct Term **);} ofc;
    } term;
} Term, *Node, **Nodes;


#define ALLOC(var) Node var = calloc(1, sizeof(Term))
#define NEW(var, ...) ALLOC(var); *var = (Term)__VA_ARGS__

void free_rec(Node t) {
    if (t == NULL) return;
    union TermU tu = t->term;
    switch (t->class) {
        case LINK: case UNIT: case COUNIT: case WEAK:
            break;
        case TENS: free_rec(tu.tens.l); free_rec(tu.tens.r); break;
        case PAR: free_rec(tu.par.l); free_rec(tu.par.r); break;
        case INL: free_rec(tu.inl); break;
        case INR: free_rec(tu.inr); break;
        case WITH: for (int i = 0; i < tu.with.len; i++)
                       free_rec(tu.with.clos[i]);
                   free(tu.with.clos);
                   break;
        case DER: free_rec(tu.der); break;
        case CONTR: free_rec(tu.contr.l); free_rec(tu.contr.r); break;
        case OFC: for (int i = 0; i < tu.ofc.len; i++)
                      free_rec(tu.ofc.clos[i]);
                  free(tu.ofc.clos);
                  break;
    }
    free(t);
}

/* constructors */

static inline void link_(Node *a, Node *b) {
    ALLOC(x); ALLOC(y);
    *x = (Term){LINK, {.link = y}};
    *y = (Term){LINK, {.link = x}};
    *a = x, *b = y;
}
#define NU(x, y) Node x, y; link_(&x, &y)
#define NU_(x) NU(x, x##_)
static inline Node unit() { NEW(t, {UNIT}); return t; }
static inline Node counit() { NEW(t, {COUNIT}); return t; }
static inline Node tens(Node l, Node r) {
    NEW(t, {TENS, {.tens = {l, r}}}); return t; }
static inline Node par(Node l, Node r) {
    NEW(t, {PAR, {.par = {l, r}}}); return t; }
static inline Node inl(Node u) { NEW(t, {INL, {.inl = u}}); return t; }
static inline Node inr(Node u) { NEW(t, {INR, {.inr = u}}); return t; }
static inline Node with(int len, Nodes clos,
        Node (*l)(Nodes), Node (*r)(Nodes)) {
    NEW(t, {WITH, {.with = {len, clos, l, r}}});
    return t;
}
static inline Node der(Node u) { NEW(t, {DER, {.der = u}}); return t; }
static inline Node weak() { NEW(t, {WEAK}); return t; }
static inline Node contr(Node l, Node r) {
    NEW(t, {CONTR, {.contr = {l, r}}}); return t; }
static inline Node ofc(int len, Nodes clos, Node (*body)(Nodes)) {
    NEW(t, {OFC, {.ofc = {len, clos, body}}});
    return t;
}

static inline Nodes copy_nodes(int len, Nodes clos) {
    Nodes new = calloc(len, sizeof(Node));
    memcpy(new, clos, len * sizeof(Node));
    return new;
}


static inline void cut(Node t, Node u) {
    Term t_val = *t, u_val = *u;
    free(t); free(u);
    union TermU tu = t_val.term, uu = u_val.term;
    if (t_val.class == LINK || u_val.class == LINK) {
        if (t_val.class == LINK) *tu.link = u_val;
        if (u_val.class == LINK) *uu.link = t_val;
    }
    /* TODO move frees to allow tail recursion */
    else do_switch: switch (t_val.class) {
        case LINK: case UNIT: case COUNIT:
            break;
        case PAR: case INL: case INR: case OFC:
            ; Term tmp_val = t_val; t_val = u_val; u_val = tmp_val;
            union TermU tmpu = tu; tu = uu; uu = tmpu;
            goto do_switch;
        case TENS:
            cut(tu.tens.l, uu.par.l);
            cut(tu.tens.r, uu.par.r);
            break;
        case WITH:
            ; Nodes clos = tu.with.clos;
            if (u_val.class == INL) cut(tu.with.l(clos), uu.inl);
            else cut(tu.with.r(clos), uu.inr);
            free(clos);
            break;
        case DER:
            clos = uu.ofc.clos; cut(uu.ofc.body(clos), tu.der);
            free(clos);
            break;
        case WEAK:
            ; int len = uu.ofc.len; clos = uu.ofc.clos;
            /* TODO optimize this into its own function, to not require
             * allocation */
            for (int i = 0; i < len; i++) cut(weak(), clos[i]);
            free(clos);
            break;
        case CONTR:
            len = uu.ofc.len; clos = uu.ofc.clos;
            Nodes clos_1 = len ? calloc(len, sizeof(Node)) : NULL,
                  clos_2 = len ? calloc(len, sizeof(Node)) : NULL;
            for (int i = 0; i < len; i++) {
                Node x_i1, x_i2;
                link_(&x_i1, clos_1 + i); link_(&x_i2, clos_2 + i);
                /* TODO optimize this into its own function, to not require
                 * allocation */
                cut(contr(x_i1, x_i2), clos[i]);
            }
            free(clos);
            cut(tu.contr.l, ofc(len, clos_1, uu.ofc.body));
            cut(tu.contr.r, ofc(len, clos_2, uu.ofc.body));
            break;
    }
}


void print_nodes(int len, Nodes a);
void print_node_prec(int prec, Node t) {
#define P(prec_, block) {if (prec > prec_) printf("("); \
    block; if (prec > prec_) printf(")");}
    switch (t->class) {
        case LINK:
            printf("%p", (void*)
                    ((size_t) t <= (size_t) t->term.link ? t : t->term.link));
            break;
        case UNIT: printf("*"); break;
        case COUNIT: printf("(*)"); break;
        case TENS: P(5, {print_node_prec(6, t->term.tens.l);
                             printf(" ⊗ ");
                             print_node_prec(5, t->term.tens.r);}); break;
        case PAR: P(4, {print_node_prec(5, t->term.par.l);
                          printf(" ⅋ ");
                          print_node_prec(4, t->term.par.r);}); break;
        case INL: printf("inl "); print_node_prec(10, t->term.inl); break;
        case INR: printf("inr "); print_node_prec(10, t->term.inr); break;
        case WITH: printf("<& box with closure [");
                   print_nodes(t->term.with.len, t->term.with.clos);
                   printf("]>"); break;
        case DER: printf("?"); print_node_prec(10, t->term.der); break;
        case WEAK: printf("_"); break;
        case CONTR: P(7, {print_node_prec(8, t->term.contr.l);
                            printf(" @ ");
                            print_node_prec(7, t->term.contr.r);}); break;
        case OFC: printf("<! box with closure [");
                  print_nodes(t->term.ofc.len, t->term.ofc.clos);
                  printf("]>"); break;
    }
}
void print_node(Node t) {print_node_prec(-1, t);}

void print_nodes(int len, Nodes a) {
    for (int i = 0; i < len; i++) {
        print_node(a[i]);
        if (i + 1 < len) printf(", ");
    }
}

int unchurch_go(Node t) {
    switch (t->class) {
        case WEAK:
            return 0;
        case CONTR:
            ; int l = unchurch_go(t->term.contr.l);
            if (l < 0) return -1;
            int r = unchurch_go(t->term.contr.r);
            if (r < 0) return -1;
            return l + r;
        case DER:
            return 1;
        default:
            return -1;
    }
}
int unchurch(Node t) {
    return unchurch_go(t->term.par.l);
}

void print_fw(Node t) {
    if (t->class == UNIT) return;
    print_fw(t->term.tens.l);
    printf("%d", t->term.tens.r->class == INL);
}


Node main_(Nodes clos);
/* paste code here generated by "putStrLn . compileToC $ somePrfExp" with
 * PE2.hs loaded, where somePrfExp :: PrfExp and has exactly 1 term in its main
 * body (if your top-level program has multiple terms in its main body, put
 * them into a big ⅋) */
/* end generated code */


int main(void) {
    Node a = main_(NULL);
    print_node(a); printf("\n");
    /* print_fw(a); printf("\n"); */
    /* printf("%d\n", unchurch(a)); */
    free_rec(a);
    return 0;
}

